;******************************************************************
; Ethernet MINI DRIVER for PIC BASIC PRO
; Author: Peter Best and Fred Eady
; Version: 1.0
; Date: 06/25/07
; Description: NUTS & VOLTS VERSION 1.0
;
; THESE ARE BASIC DRIVER ROUTINES AND ARE NOT INTENDED TO BE USED IN
; A PRODUCTION ENVIRONMENT. EDTP OFFERS THESE DRIVER ROUTINES AS-IS
; AND IS NOT RESPONSIBLE FOR ANY MISUSE OR MISAPPLICATION OF THIS 
; CODE.
;
; 7/24/2007		Brought up ARP,ICMP and UDP for EDTP test panel
;******************************************************************

include "mini.inc"
;******************************************************************
;*	GENERAL GLOBAL VARIABLES
;******************************************************************
tx_cnt	var word
tx_end	var word
tx_status	var byte[6]
hdrlen		var word
chksum16	var word
acc3		var word
acc2		var word
acc1		var word
acc0		var word
arg3		var byte
arg2		var byte
arg1		var byte
arg0		var byte
i16			var word
i8			var byte
data8		var byte
data16		var word
rgstr		var byte
low_byte	var byte
high_byte	var byte
temp		var byte
forever 	con 1
;******************************************************************
;*	FRAME HANDLER VARIABLES
;******************************************************************
bufferptr		var byte
txlen			var word	
rxlen			var word
new_rdptr 		var word
;******************************************************************
;*	ARP DEFINITIONS
;******************************************************************
i_arp	var word
;******************************************************************
;*	ICMP DEFINITIONS
;******************************************************************
i_icmp	var word
;******************************************************************
;*	UDP DEFINITIONS
;******************************************************************
i_udp	var word
;******************************************************************
;*	DHCP Definitions
;******************************************************************
i_dhcp		var word
DHCP_CLIENT_PORT					CON	$44
DHCP_SERVER_PORT					CON	$43

BOOT_REQUEST    					CON	$01
BOOT_REPLY      					CON	$02
HW_TYPE         					CON	$01
HW_TYPE_LEN							CON	$06

DHCP_MESSAGE_TYPE					CON	$35
DHCP_MESSAGE_TYPE_LEN				CON	$01

DHCP_UNKNOWN_MESSAGE				CON	$00

DHCP_DISCOVER_MESSAGE           	CON	$01
DHCP_OFFER_MESSAGE					CON	$02
DHCP_REQUEST_MESSAGE				CON	$03
DHCP_DECLINE_MESSAGE 				CON	$04
DHCP_ACK_MESSAGE					CON	$05
DHCP_NAK_MESSAGE					CON	$06
DHCP_RELEASE_MESSAGE            	CON	$07

DHCP_SERVER_IDENTIFIER          	CON	$36
DHCP_SERVER_IDENTIFIER_LEN      	CON	$04

DHCP_PARAM_REQUEST_LIST         	CON	$37
DHCP_PARAM_REQUEST_LIST_LEN     	CON	$02
DHCP_PARAM_REQUEST_IP_ADDRESS       CON	$32
DHCP_PARAM_REQUEST_IP_ADDRESS_LEN   CON	$04
DHCP_SUBNET_MASK					CON	$01
DHCP_ROUTER     					CON	$03
DHCP_IP_LEASE_TIME              	CON	$33
DHCP_END_OPTION 					CON	$FF
;******************************************************************
;*	DHCP STATES
;******************************************************************
DHCPSTATE	VAR	byte
DHCP_ENTRY		CON	$0					
DHCP_INIT		CON	$1						
DHCP_WAIT		CON	$2
DHCP_BROADCAST	CON	$3
DHCP_DISCOVER	CON	$4
DHCP_REQUEST	CON	$5
DHCP_BIND		CON	$6
DHCP_BOUND		CON	$7
DHCP_DISABLED	CON	$8
;******************************************************************
;*	ETHERNET BUFFER DEFINITIONS 
;******************************************************************
MAXFRAME		CON	1518
TX_BUFFER_SIZE	CON	1518
TXEND			CON TX_BUFFER_SIZE + 8
TXSTART			CON	8192 - TXEND
RXSTART			CON	$0000
RXSTOPMOD		CON TXSTART - 2
RXSTOP			CON	RXSTOPMOD | $0001
RXSIZEMOD		CON	RXSTART + 1
RXSIZE			CON	RXSTOP-RXSIZEMOD
;******************************************************************
;*	IP ADDRESS DEFINITION
;*   These are the default Ethernet Module IP addresses.
;*   You may change these to any valid aDEFINEress.
;******************************************************************
ipaddrascii 				var byte[3]
ipaddrc  					var byte[4]
ipaddrc[0] = 192
ipaddrc[1] = 168
ipaddrc[2] = 0
ipaddrc[3] = 150
tempipaddrc 				var byte[4]
svridc 						var byte[4]
gwayipaddrc 				var byte[4]
tempgwayipaddrc 			var byte[4]
subnetmaskc 				var byte[4]
tempsubnetmaskc 			var byte[4]
tempsvridc 					var byte[4]
templeasetimec 				var byte[4]
broadcastipaddrc  			var byte[4]
broadcastipaddrc[0] = 192
broadcastipaddrc[1] = 168
broadcastipaddrc[2] = 0
broadcastipaddrc[3] = 255
;******************************************************************
;*	HARDWARE (MAC) ADDRESS DEFINITION
;*   This is the default Ethernet Module hardware address.
;*   You may change this to any valid address.
;******************************************************************
macaddrc		var	byte[6]
macaddrc[0] = 0
macaddrc[1] = 0
macaddrc[2] = "E"
macaddrc[3] = "D"
macaddrc[4] = "T"
macaddrc[5] = "P"
remotemacaddrc	var byte[6] 
svrmacaddrc		var byte[6]
;******************************************************************
;*	Receive Ring Buffer Header Layout
;*   This is the 6-byte header that resides infront of the
;*   data packet in the receive buffer.
;******************************************************************
tx_control_byte	   	CON	$0E
packetheader	var byte[6]
nextpacket_low    	CON	$00
nextpacket_high   	CON	$01
rec_bytecnt_low    	CON	$02
rec_bytecnt_high   	CON	$03
rec_status_low 	   	CON	$04
rec_status_high	   	CON	$05
;******************************************************************
;*	Ethernet Header Layout
;******************************************************************
udp_packet	var byte[100]                   
packet		var byte[1500]                
enetpacketDest0	   	CON	$00  ;destination mac address
enetpacketDest1	   	CON	$01
enetpacketDest2	   	CON	$02
enetpacketDest3	   	CON	$03
enetpacketDest4	   	CON	$04
enetpacketDest5	   	CON	$05
enetpacketSrc0	   	CON	$06  ;source mac address
enetpacketSrc1	   	CON	$07
enetpacketSrc2	   	CON	$08
enetpacketSrc3	   	CON	$09
enetpacketSrc4	   	CON	$0A
enetpacketSrc5	   	CON	$0B
enetpacketType0	   	CON	$0C  ;type/length field
enetpacketType1	   	CON	$0D
enetpacketData     	CON	$0E  ;IP data area begins here
;******************************************************************
;*	ARP Layout
;******************************************************************
arp_hwtype		   CON	$0E
arp_prtype		   CON	$10
arp_hwlen		   CON	$12
arp_prlen		   CON	$13
arp_op			   CON	$14	  ;arp request or response
arp_shaddr		   CON	$16   ;arp source mac address
arp_sipaddr		   CON	$1C   ;arp source ip address
arp_thaddr		   CON	$20   ;arp target mac address
arp_tipaddr		   CON	$26   ;arp target ip address
;******************************************************************
;*	IP Header Layout
;******************************************************************
ip_vers_len		   CON	$0E	;IP version and header length 1a19
ip_tos			   CON	$0F	;IP type of service
ip_pktlen		   CON	$10	;packet length
ip_id			   CON	$12	;datagram id
ip_frag_offset	   CON	$14	;fragment offset
ip_ttl			   CON	$16	;time to live
ip_proto		   CON	$17	;protocol (ICMP=1, TCP=6, UDP=11)
ip_hdr_cksum	   CON	$18	;header checksum 1a23
ip_srcaddr		   CON	$1A	;IP address of source
ip_destaddr		   CON	$1E	;IP addess of destination
ip_data			   CON	$22	;IP data area
;******************************************************************
;*	TCP Header Layout
;******************************************************************
TCP_srcport		   CON	$22	;TCP source port
TCP_destport   	   CON	$24	;TCP destination port
TCP_seqnum  	   CON	$26	;sequence number
TCP_acknum	       CON	$2A	;acknowledgement number
TCP_hdrflags	   CON	$2E	;4-bit header len and flags
TCP_window		   CON	$30	;window size
TCP_cksum		   CON	$32	;TCP checksum
TCP_urgentptr      CON	$34	;urgent pointer
TCP_data           CON	$36 ;option/data
;******************************************************************
;*	TCP Flags
;*   IN flags represent incoming bits
;*   OUT flags represent outgoing bits
;******************************************************************
TCPFLAGS var packet[TCP_hdrflags+1]
FIN_IN               CON  $01
SYN_IN               CON  $02
RST_IN               CON  $04
PSH_IN               CON  $08
ACK_IN               CON  $10
URG_IN               CON  $20
;FIN_OUT              CON packet[TCP_hdrflags+1] |= $01 ;0000 0001 
;SYN_OUT              CON packet[TCP_hdrflags+1] |= $02 ;0000 0010 
;RST_OUT              CON packet[TCP_hdrflags+1] |= $04 ;0000 0100 
;PSH_OUT              CON packet[TCP_hdrflags+1] |= $08 ;0000 1000 
;ACK_OUT              CON packet[TCP_hdrflags+1] |= $10 ;0001 0000 
;URG_OUT              CON packet[TCP_hdrflags+1] |= $20 ;0010 0000 
;******************************************************************
;*	IP Protocol Types
;******************************************************************
PROT_ICMP			   CON $01
PROT_TCP			   CON $06
PROT_UDP			   CON $11
;******************************************************************
;*	ICMP Header
;******************************************************************
ICMP_type			   CON ip_data
ICMP_code			   CON ICMP_type+1
ICMP_cksum			   CON ICMP_code+1
ICMP_id				   CON ICMP_cksum+2
ICMP_seqnum			   CON ICMP_id+2
ICMP_data              CON ICMP_seqnum+2
;******************************************************************
;*	UDP Header
;******************************************************************
UDP_srcport			   CON ip_data
UDP_destport		   CON UDP_srcport+2
UDP_len				   CON UDP_destport+2
UDP_cksum			   CON UDP_len+2
UDP_data			   CON UDP_cksum+2

LED0			var LATF.1
;TIMER VARIABLES
milliseconds	var word
msecs_timer 	var word
clock_timer 	var word
hours			var byte
mins			var byte
secs			var byte
secs_timer		var byte
;EUSART VARIABLES
tmphead			var byte
tmptail			var byte
returncode		var word
recvchar_ret	var byte
sendchar_ret	var byte
CharInQueue_ret	var byte
;*****************************************************************
;*	USART Definitions
;******************************************************************
BAUD	CON		57600		;desired baud rate
DEFINE OSC		25			;oscillator frequency
DIVIDER CON 	43	;81			;((unsigned int)(FOSC/(16 * BAUD) -1))
;1,2,4,8,16,32,64,128 or 256 bytes 
USART_RX_BUFFER_SIZE  CON 256      
USART_RX_BUFFER_MASK CON  USART_RX_BUFFER_SIZE - 1 
; 1,2,4,8,16,32,64,128 or 256 bytes 
USART_TX_BUFFER_SIZE  CON 256      
USART_TX_BUFFER_MASK CON USART_TX_BUFFER_SIZE - 1 

USART_RxBuf		var	byte[USART_RX_BUFFER_SIZE]
USART_TxBuf		var byte[USART_TX_BUFFER_SIZE]
USART_TxHead	var byte
USART_TxTail	var byte
USART_RxHead	var byte
USART_RxTail	var byte
;******************************************************************
;*	MAIN PROGRAM
;******************************************************************
on interrupt goto int_handler

	OSCTUNE = $40
;******************************************************************
;*	CONFIGURE AND START TIMER2
;* SET TO OVERFLOW EVERY 1mS
;******************************************************************
 	hours = 12
	mins = $00 
	secs = $00 
  	milliseconds = 0
	T2CON = $79 
	PR2 = $A3 		
	TMR2ON = 1
	TMR2IE = 1
	gosub init_EUSART
	gosub init_IO
	gosub init_67J60

	while forever
	 if EPKTCNT != 0  then
		gosub get_frame
	 endif
	wend
end
;******************************************************************
;*	INTERRUPT HANDLER
;******************************************************************
	Disable
int_handler:
	if(TMR2IF == 1 && TMR2IE == 1) then
   		TMR2IF = 0
        msecs_timer = msecs_timer + 1
		clock_timer = clock_timer + 1
 
	if(1000 == clock_timer) then
			clock_timer = 0
			secs = secs + 1
            secs_timer = secs_timer + 1
		    ;lease_timer = lease_timer + 1
			toggle LED0
	endif
  
	if(60 == secs) then
			secs = 0
		 	mins = mins + 1
	endif

	if(60 == mins) then
			mins = 0
			hours = hours + 1
	endif

	if(24 == hours) then
		    hours = 0
	endif	
endif
	
  	if(RC1IF == 1) then
   		data8 = RCREG1                   ; read the received data 
                                		; calculate buffer index 
   		tmphead = ( USART_RxHead + 1 ) & USART_RX_BUFFER_MASK
   		USART_RxHead = tmphead        ; store new index 
  	endif

   	if ( tmphead == USART_RxTail ) then
     ; ERROR! Receive buffer overflow 
	endif                                
   USART_RxBuf[tmphead] = data8   ; store received data in buffer 
  
  if(TRMT == 1) then
                                ; check if all data is transmitted  
	if ( USART_TxHead != USART_TxTail ) then
                                ; calculate buffer index  
     tmptail = ( USART_TxTail + 1 ) & USART_TX_BUFFER_MASK
     USART_TxTail = tmptail      ; store new index  

    TXREG1 = USART_TxBuf[tmptail]  ; start transmition  
   
   	else
   
     TX1IE = 0         ; disable TX interrupt  
	endif
  endif
	resume
	enable

checkme:
hserout["A"]
pause 1
return
;******************************************************************
;*	Init USART Function
;******************************************************************
init_EUSART:
  	BRGH = 1 
  	SYNC = 0
  	SPBRG1 = DIVIDER 	;load baud rate divisor	
  	TRISC7 = 1 			;receive pin
  	TRISC6 = 1 			;transmit pin
  	TXSTA1 = $04 		;high speed baud rate
  	RCSTA1 = $80 		;enable serial port and serial port pins

  
  	USART_RxTail = $00 	;flush receive buffer 
  	USART_RxHead = $00 
  	USART_TxTail = $00 	;flush transmit buffer
  	USART_TxHead = $00 
  	RC1IP = 1 			;receive interrupt = high priority
  	TX1IP = 1 			;transmit interrupt = high priority
  	RC1IE = 1 			;enable receive interrupt
  	PEIE = 1 				;enable all unmasked peripheral interrupts
  	GIE = 1 				;enable all unmasked interrupts
  	CREN = 1 				;enable USART1 receiver
  	TX1IE = 0 			;disable USART1 transmit interrupt
  	TXEN = 1 				;transmitter enabled
  	SPEN = 1 
	hserout[13,10]
	hserout[13,10,"EUSART Initialized"]
	pause 10
	return
recvchar:
  ; wait for incomming data
  
	while  USART_RxHead == USART_RxTail  
	wend
                                ; calculate buffer index  
  	tmptail = ( USART_RxTail + 1 ) & USART_RX_BUFFER_MASK
  	USART_RxTail = tmptail        ; store new index  
  	recvchar_ret = USART_RxBuf[tmptail]   ; return data  
  	return 

sendchar:
    ; calculate buffer index 
  	tmphead = ( USART_TxHead + 1 ) & USART_TX_BUFFER_MASK
                                ; wait for free space in buffer 
	while  tmphead == USART_TxTail  
	wend
                                ; store data in buffer 
  	USART_TxBuf[tmphead] = data8;
  	USART_TxHead = tmphead;        ; store new index 

  	TX1IE = 1            ; enable TX interrupt 
  	sendchar_ret = data8
  	return 

CharInQueue:
  	if(USART_RxHead != USART_RxTail) then
  		CharInQueue_ret = 1
  	else
  		CharInQueue_ret = 0
  	endif
  	return 
init_IO:
	TRISA = %11111100 
	TRISB = %11111111 
	TRISC = %01010111 
	TRISD = %11111011 
	TRISF = $00 
	LATF = $FF 
	hserout[13,10,"I/O Initialized"]
	pause 10
	return
init_67J60:
	;Enable Ethernet
	ETHEN = 1 

	while !PHYRDY
	wend

	msecs_timer = 0 
 	while msecs_timer < 2
@	nop
 	wend
 
	LATF = $00 
	
	RXEN = 0 
   	TXRTS = 0 

	packetheader[nextpacket_low] = RXSTART & $00FF 			;LOW_BYTE(RXSTART) 
	packetheader[nextpacket_high] = (RXSTART & $FF00) >> 8 	;HIGH_BYTE(RXSTART) 

	ERXSTL = RXSTART & $00FF		;LOW_BYTE(RXSTART) 
	ERXSTH = (RXSTART & $FF00) >> 8 	;HIGH_BYTE(RXSTART) 
	ERXRDPTL = RXSTOP & $00FF 		;LOW_BYTE(RXSTOP) 
	ERXRDPTH = (RXSTOP & $FF00) >> 8 ;HIGH_BYTE(RXSTOP) 
	ERXNDL = RXSTOP & $00FF 		;LOW_BYTE(RXSTOP) 
	ERXNDH = (RXSTOP & $FF00) >> 8 	;HIGH_BYTE(RXSTOP) 
	ETXSTL =  TXSTART & $00FF 		;LOW_BYTE(TXSTART) 
	ETXSTH =  (TXSTART & $FF00) >> 8 ;HIGH_BYTE(TXSTART) 

	MACON1 =  MACON1_TXPAUS | MACON1_RXPAUS | MACON1_MARXEN 
@	nop
	; Pad packets to 60 bytes, add CRC, and check Type/Length field
	MACON3 =  MACON3_PADCFG0 | MACON3_TXCRCEN | MACON3_FRMLNEN 
@	nop
    ; Allow infinite deferals if the medium is continuously busy 
    ; (do not time out a transmission if the half duplex medium is 
    ; completely saturated with other people's data)
    MACON4 =  MACON4_DEFER 
@	nop

	MAIPGL =  $12 
@	nop
	MAIPGH =  $0C 
@	nop
	MAMXFLL =  MAXFRAME & $00FF 		;LOW_BYTE(MAXFRAME) 	
@	nop
	MAMXFLH =  (MAXFRAME & $FF00) >> 8 	;HIGH_BYTE(MAXFRAME) 
@	nop

    MAADR1 =  macaddrc[0] 
@	nop
    MAADR2 =  macaddrc[1] 
@	nop
    MAADR3 =  macaddrc[2] 
@	nop
   MAADR4 =  macaddrc[3] 
@	nop
    MAADR5 =  macaddrc[4] 
@	nop
    MAADR6 =  macaddrc[5] 
@	nop

	rgstr = PHCON2
	data16 = $0110
	gosub wr_phy		;wr_phy(PHCON2, $0110) 
	rgstr = PHCON1
	data16 = $0000
	gosub wr_phy		;wr_phy(PHCON1,$0000) 
	rgstr = PHLCON
	data16 = $0472
	gosub wr_phy		;wr_phy(PHLCON, $0472) 	

	; Set the back-to-back inter-packet gap time to IEEE specified 
	; requirements.  The meaning of the MABBIPG value changes with the duplex
	; state, so it must be updated in this function.
	; In full duplex, $15 represents 9.6us  $12 is 9.6us in half duplex
	
	MABBIPG = $12 	

	; Enable packet reception
	RXEN =1 
    EIE = $00 
    
;	FOR TEST ONLY
;	rgstr = PHCON2 
;	gosub rd_phy
;@	nop 
;	rgstr = PHCON1 
;	gosub rd_phy
;@	nop 
;	rgstr = PHLCON 
;	gosub rd_phy
;@	nop 
   
	hserout[13,10,"Ethernet Initialized",13,10] 
	RETURN
;******************************************************************
;*	rd_phy
;******************************************************************
rd_phy:
	; Set the right address and start the register read operation
	MIREGADR = rgstr
@	nop
	MICMD = MICMD_MIIRD
@	nop
	; Loop to wait until the PHY register has been read through the MII
	while BUSY
	wend
	; Stop reading
	MICMD = $00
@	nop
	; Obtain results and return
	low_byte = MIRDL
	high_byte = MIRDH
	returncode = (high_byte << 8) | low_byte
	return
;******************************************************************
;*	wr_phy
;******************************************************************
wr_phy:
	; Write the register address
	MIREGADR = rgstr;
@	nop
	; Write the data
	; Order is important: write low byte first, high byte last
	MIWRL = data16 & $00FF 	;LOW_BYTE(data)
@	nop
	MIWRH = (data16 & $FF00) >> 8	;HIGH_BYTE(data 
@	nop
	; Wait until the PHY register has been written
	while BUSY
	wend
	return
;******************************************************************
;*	Get A Frame From the Buffer
;*   This routine removes an Ethernet frame from the receive buffer
;*   ring.
;******************************************************************
get_frame:

	;banksel(ERDPTL 
	ERDPTL = packetheader[nextpacket_low]
	ERDPTH = packetheader[nextpacket_high]

	bufferptr = -1
	i16 = 0
	while i16 < 6
		bufferptr = bufferptr + 1
		i16 = i16 + 1
		packetheader[bufferptr] = EDATA;
	wend
 
	rxlen = (packetheader[rec_bytecnt_high] << 8) | packetheader[rec_bytecnt_low]

	bufferptr = -1
	i16 = rxlen;
	while i16
		bufferptr = bufferptr + 1
		i16 = i16 - 1
		packet[bufferptr] = EDATA;
	wend
	
	new_rdptr = ((packetheader[nextpacket_high] << 8) | packetheader[nextpacket_low]) - 1
	if new_rdptr > 8191 || new_rdptr > RXSTOP then
		new_rdptr = RXSTOP
	endif
	 
	ECON2 = ECON2 | ECON2_PKTDEC
	
	;banksel(ERXRDPTL 
	ERXRDPTL = new_rdptr & $00FF		;LOW_BYTE(new_rdptr 
	ERXRDPTH = (new_rdptr & $FF00) >> 8	;HIGH_BYTE(new_rdptr 
	
   	;process an incoming ARP request packet
   	if packet[enetpacketType0] == $08 && packet[enetpacketType1] == $06  then
	   if packet[arp_hwtype+1] == $01 		&&_
       	packet[arp_prtype] == $08 			&&_ 
	   	packet[arp_prtype+1] == $00 		&&_
       	packet[arp_hwlen] == $06 			&&_ 
	   	packet[arp_prlen] == $04 			&&_
       	packet[arp_tipaddr]   == ipaddrc[0] &&_
       	packet[arp_tipaddr+1] == ipaddrc[1] &&_
       	packet[arp_tipaddr+2] == ipaddrc[2] &&_
       	packet[arp_tipaddr+3] == ipaddrc[3] then

		select case packet[arp_op+1]
			case $01
				gosub arp_reply
			case $02
       			for i8 = 0 to 5
		 		 remotemacaddrc[i8] = packet[arp_shaddr+i8] 
				next i8
				gosub set_arpflag
		end select		
	   endif	
	 endif

   	;process an IP packet
   		if packet[enetpacketType0] == $08      &&_ 
		  packet[enetpacketType1] == $00 	   &&_
          packet[ip_destaddr]   == ipaddrc[0]  &&_
          packet[ip_destaddr+1] == ipaddrc[1]  &&_
          packet[ip_destaddr+2] == ipaddrc[2]  &&_
          packet[ip_destaddr+3] == ipaddrc[3]  then
	
      	select case packet[ip_proto]
	      case PROT_ICMP
	      	gosub icmp
	      case PROT_UDP
    	  	if packet[UDP_srcport] == DHCP_SERVER_PORT then
    	  	 	gosub dhcp_state_engine
    	  	else
    	  	 	gosub udp
    	  	endif
    	  case PROT_TCP
    	  	 gosub tcp
    	  ;case else
	      ;	LED0 = 1
		 end select
		endif
  	;process a DHCP message
   	 	if packet[enetpacketType0] == $08	&&_ 
		  packet[enetpacketType1] == $00 	&&_
          packet[ip_destaddr]   == $FF  	&&_
          packet[ip_destaddr+1] == $FF  	&&_
          packet[ip_destaddr+2] == $FF  	&&_
          packet[ip_destaddr+3] == $FF	 	&&_
          packet[ip_proto] == PROT_UDP then
		  gosub set_newdhcppkt
    	  gosub dhcp_state_engine
	 	endif
	return

disable_TMR2int:
  	TMR2IE = 0
return
enable_TMR2int:
	TMR2IE = 1
return
disable_TMR3int:
	TMR3IE = 0
return
enable_TMR3int:
	TMR3IE = 1
return

enable_GLOBALint:
	IPEN = 0
	GIE = 1				
    PEIE = 1
return
disable_GLOBALint:
	IPEN = 0
	GIE = 0				
    PEIE = 0
return
;******************************************************************
;*	Flags
;******************************************************************
miniflags var byte
synflag 	CON	$01  
finflag		CON	$02  
entryflag	CON	$04
arpflag		CON	$08
hexflag		CON	$10
bsynflag:   rc1 = 0: if miniflags & synflag	then rc1 = 1	: return
bfinflag:	rc1 = 0: if miniflags & finflag	then rc1 = 1	: return
bentryflag:	rc1 = 0: if miniflags & entryflag then rc1 = 1	: return
barpflag:	rc1 = 0: if miniflags & arpflag	then rc1 = 1	: return
bhexflag:	rc1 = 0: if miniflags & hexflag	then rc1 = 1	: return

clr_synflag: 		miniflags = miniflags & ~synflag	: return
set_synflag: 		miniflags = miniflags | synflag		: return
clr_finflag: 		miniflags = miniflags & ~finflag	: return
set_finflag: 		miniflags = miniflags | finflag		: return
clr_entryflag: 		miniflags = miniflags & ~entryflag	: return
set_entryflag: 		miniflags = miniflags | entryflag	: return
clr_arpflag: 		miniflags = miniflags & ~arpflag	: return
set_arpflag: 		miniflags = miniflags | arpflag		: return
clr_hexflag: 		miniflags = miniflags & ~hexflag	: return
set_hexflag: 		miniflags = miniflags | hexflag		: return

dhcpflags var byte
bound		CON	$01  
offerrec	CON	$02  
done		CON	$04
trashit		CON $08
newdhcppkt	CON	$10
bbound:			rc1 = 0: if dhcpflags & bound then rc1 = 1 		: return
bofferrec:		rc1 = 0: if dhcpflags & offerrec then rc1 = 1 	: return
bdone:			rc1 = 0: if dhcpflags & done then rc1 = 1 		: return
btrashit:		rc1 = 0: if dhcpflags & trashit then rc1 = 1 	: return
bnewdhcppkt:	rc1 = 0: if dhcpflags & newdhcppkt then rc1 = 1 : return

clr_bound:		dhcpflags = dhcpflags & ~bound		: return	
set_bound: 		dhcpflags = dhcpflags | bound		: return
clr_offerrec: 	dhcpflags = dhcpflags & ~offerrec	: return
set_offerrec: 	dhcpflags = dhcpflags | offerrec	: return
clr_done:		dhcpflags = dhcpflags & ~done		: return
set_done: 		dhcpflags = dhcpflags | done		: return
clr_trashit:	dhcpflags = dhcpflags & ~trashit	: return
set_trashit:	dhcpflags = dhcpflags | trashit		: return
clr_newdhcppkt:	dhcpflags = dhcpflags & ~newdhcppkt	: return
set_newdhcppkt:	dhcpflags = dhcpflags | newdhcppkt	: return

arp_reply:
;******************************************************************
;*	Perform ARP Response
;*   This routine supplies a requesting computer with the
;*   Ethernet modules's MAC (hardware) address.
;******************************************************************

	tx_end = TXSTART;
   ;load beginning page for transmit buffer
    ;banksel(EWRPTL 
	EWRPTL = TXSTART & $00FF
	EWRPTH = (TXSTART & $FF00) >> 8

   ;write control byte
	EDATA = tx_control_byte 
	tx_end = tx_end + 1

   ;write destination MAC address
   for i_arp = 0 to 5
      EDATA = packet[enetpacketSrc0+i_arp] 
	  tx_end = tx_end + 1
	next i_arp
	  
   ;write source MAC address
   for i_arp = 0 to 5
      EDATA = macaddrc[i_arp] 
	  tx_end = tx_end + 1
	next i_arp

   ;write typelen hwtype prtype hwlen prlen op:
   ;addr = &packet[enetpacketType0]
   packet[arp_op+1] = $02
   for i_arp = 0 to 9
      EDATA = packet[enetpacketType0+i_arp] 
	  tx_end = tx_end + 1
	next i_arp
   ;write ethernet module MAC address
   for i_arp = 0 to 5
      EDATA = macaddrc[i_arp] 
	  tx_end = tx_end + 1
	next i_arp
   ;write ethernet module IP address
      for i_arp = 0 to 3
      	 EDATA = ipaddrc[i_arp] 
	     tx_end = tx_end + 1
	next i_arp

   ;write remote MAC address
   for i_arp = 0 to 5
      EDATA = packet[enetpacketSrc0+i_arp] 
	  tx_end = tx_end + 1
	next i_arp

   ;write remote IP address
   for i_arp = 0 to 3
      EDATA = packet[arp_sipaddr+i_arp] 
	  tx_end = tx_end + 1
	next i_arp
	;banksel(ETXNDL 
	ETXNDL = tx_end & $00FF 
	ETXNDH = (tx_end & $FF00) >> 8 

   ;send the contents of the transmit buffer onto the network
	if TXERIF then
	 TXERIF = 0
     TXRST = 1
   	 TXRST = 0
	endif
	TXIF = 0
   	TXRTS = 1
	while TXRTS
	wend 
	tx_end = tx_end + 1
	;banksel(ERDPTL 
	ERDPTL= tx_end & $00FF 
	ERDPTH= (tx_end & $FF00) >> 8 

	;banksel(ERDPTL 
    for i_arp = 0 to 6	
		tx_status[i_arp]= EDATA 
	next i_arp
	if TXABRT then
	   hserout["13,10,ARP Reply Transmission Aborted.."] 
	endif
return

icmp:
;******************************************************************
;*	Perform ICMP Function
;*   This routine responds to a ping.
;******************************************************************
   ;set echo reply
   packet[ICMP_type]=$00
   packet[ICMP_code]=$00

   ;clear the ICMP checksum
   packet[ICMP_cksum]=$00
   packet[ICMP_cksum+1]=$00

   ;setup the IP header
   gosub setipaddrs

   hdrlen = ((packet[ip_pktlen] << 8) + packet[ip_pktlen+1]) - ((packet[ip_vers_len] & $0F) * 4)
   ;addr = &packet[ICMP_type]
	i_icmp = 0
	gosub clrcrc
	while(hdrlen > 1)
		arg1=packet[ICMP_type+i_icmp]
		i_icmp = i_icmp + 1
		arg0=packet[ICMP_type+i_icmp]
		i_icmp = i_icmp + 1
		acc0 = acc0 + arg0
		gosub ripple_crc
		acc1 = acc1 + arg1
		gosub ripple_crc
		hdrlen = hdrlen - 2
	wend
      if(hdrlen > 0) then
        arg1=packet[ICMP_type+i_icmp]
        arg0=$00
		acc0 = acc0 + arg0
		gosub ripple_crc
		acc1 = acc1 + arg1
		gosub ripple_crc
	 endif
		acc0 = acc0 + acc2
		gosub ripple_crc
		acc1 = acc1 + acc3
		gosub ripple_crc
   chksum16= ~((acc1 << 8) + acc0);
   packet[ICMP_cksum] = (chksum16 & $FF00) >> 8
   packet[ICMP_cksum+1] = chksum16 & $00FF
   ;send the ICMP packet along on its way
	tx_end = TXSTART
   ;load beginning page for transmit buffer
    ;banksel(EWRPTL);
	EWRPTL = TXSTART & $00FF
	EWRPTH = (TXSTART & $FF00) >> 8

   ;write control byte
	EDATA = tx_control_byte
	tx_end = tx_end + 1

	tx_end = (tx_end) + (packetheader[rec_bytecnt_high] << 8) + (packetheader[rec_bytecnt_low] - 4);;4 is checksum of incoming packet
	;banksel(ETXNDL);
	ETXNDL = tx_end & $00FF 
	ETXNDH = (tx_end & $FF00) >> 8 
	
	i_icmp = 0
	while(tx_end)
		EDATA = packet[0 + i_icmp]
		i_icmp = i_icmp + 1
		tx_end = tx_end - 1
	wend		

   ;send the contents of the transmit buffer onto the network
	if TXERIF then
	 TXERIF = 0 
     TXRST = 1 
   	 TXRST = 0 
	endif
  	TXRTS = 1
	while(1 == (ECON1 & ECON1_TXRTS))
	wend
return
setipaddrs:
;******************************************************************
;*	SETIPADDRS
;*   This subroutine builds the IP header.
;******************************************************************
   ;move IP source address to destination address
   packet[ip_destaddr]=packet[ip_srcaddr]
   packet[ip_destaddr+1]=packet[ip_srcaddr+1]
   packet[ip_destaddr+2]=packet[ip_srcaddr+2]
   packet[ip_destaddr+3]=packet[ip_srcaddr+3]
   ;make ethernet module IP address source address
   packet[ip_srcaddr]=ipaddrc[0]
   packet[ip_srcaddr+1]=ipaddrc[1]
   packet[ip_srcaddr+2]=ipaddrc[2]
   packet[ip_srcaddr+3]=ipaddrc[3]
   ;move hardware source address to destinatin address
   packet[enetpacketDest0]=packet[enetpacketSrc0]
   packet[enetpacketDest1]=packet[enetpacketSrc1]
   packet[enetpacketDest2]=packet[enetpacketSrc2]
   packet[enetpacketDest3]=packet[enetpacketSrc3]
   packet[enetpacketDest4]=packet[enetpacketSrc4]
   packet[enetpacketDest5]=packet[enetpacketSrc5]
   ;make ethernet module mac address the source address
   packet[enetpacketSrc0]=macaddrc[0]
   packet[enetpacketSrc1]=macaddrc[1]
   packet[enetpacketSrc2]=macaddrc[2]
   packet[enetpacketSrc3]=macaddrc[3]
   packet[enetpacketSrc4]=macaddrc[4]
   packet[enetpacketSrc5]=macaddrc[5]

   ;calculate the IP header checksum
   packet[ip_hdr_cksum]=$00
   packet[ip_hdr_cksum+1]=$00

   hdrlen = (packet[ip_vers_len] & $0F) * 4
   ;addr = &packet[ip_vers_len]
	i16 = 0
	gosub clrcrc
	while(hdrlen > 1)
		arg1=packet[ip_vers_len+i16]
		i16 = i16 + 1
		arg0=packet[ip_vers_len+i16]
		i16 = i16 + 1
		acc0 = acc0 + arg0
		gosub ripple_crc
		acc1 = acc1 + arg1
		gosub ripple_crc
		hdrlen = hdrlen - 2
	wend
      if(hdrlen > 0) then
        arg1=packet[ip_vers_len+i16]
        arg0=$00
		acc0 = acc0 + arg0
		gosub ripple_crc
		acc1 = acc1 + arg1
		gosub ripple_crc
	 endif
		acc0 = acc0 + acc2
		gosub ripple_crc
		acc1 = acc1 + acc3
		gosub ripple_crc
   chksum16= ~((acc1 << 8) + acc0);
   packet[ip_hdr_cksum] = (chksum16 & $FF00) >> 8
   packet[ip_hdr_cksum+1] = chksum16 & $00FF
return


udp:
;******************************************************************
;*	UDP Function
;*   This function uses a Visual Basic UDP program to echo the
;*   data back to the VB program .
;******************************************************************
	;unsigned int buffer_addr,i,tx_end;	
;	char *bufferptr,temp;
	
   ;port 7 is the well-known echo port
   if(packet[UDP_destport] == $00 && packet[UDP_destport+1] ==$07) then
      ;build the IP header
      gosub setipaddrs

      ;swap the UDP source and destination ports
      temp = packet[UDP_srcport]
      packet[UDP_srcport] = packet[UDP_destport]
      packet[UDP_destport] = temp;

      temp = packet[UDP_srcport+1]
      packet[UDP_srcport+1] = packet[UDP_destport+1]
      packet[UDP_destport+1] = temp;

      ;calculate the UDP checksum
      packet[UDP_cksum] = $00;
      packet[UDP_cksum+1] = $00;

      hdrlen = $08;
      ;addr = &packet[ip_srcaddr]
	i16 = 0
	gosub clrcrc
	while(hdrlen > 1)
		arg1=packet[ip_srcaddr+i16]
		i16 = i16 + 1
		arg0=packet[ip_srcaddr+i16]
		i16 = i16 + 1
		acc0 = acc0 + arg0
		gosub ripple_crc
		acc1 = acc1 + arg1
		gosub ripple_crc
		hdrlen = hdrlen - 2
	wend
      if(hdrlen > 0) then
        arg1=packet[ip_srcaddr+i16]
        arg0=$00
		acc0 = acc0 + arg0
		gosub ripple_crc
		acc1 = acc1 + arg1
		gosub ripple_crc
	 endif

		acc0 = acc0 + packet[ip_proto]
		gosub ripple_crc

		acc0 = acc0 + packet[UDP_len+1]
		gosub ripple_crc
		acc1 = acc1 + packet[UDP_len]
		gosub ripple_crc

	hdrlen = (packet[UDP_len] << 8) + packet[UDP_len+1]
	i16 = 0
	while(hdrlen > 1)
		arg1=packet[UDP_srcport+i16]
		i16 = i16 + 1
		arg0=packet[UDP_srcport+i16]
		i16 = i16 + 1
		acc0 = acc0 + arg0
		gosub ripple_crc
		acc1 = acc1 + arg1
		gosub ripple_crc
		hdrlen = hdrlen - 2
	wend
      if(hdrlen > 0) then
        arg1=packet[UDP_srcport+i16]
        arg0=$00
		acc0 = acc0 + arg0
		gosub ripple_crc
		acc1 = acc1 + arg1
		gosub ripple_crc
	 endif

		acc0 = acc0 + acc2
		gosub ripple_crc
		acc1 = acc1 + acc3
		gosub ripple_crc

   chksum16= ~((acc1 << 8) + acc0);
   packet[UDP_cksum] = (chksum16 & $FF00) >> 8
   packet[UDP_cksum+1] = chksum16 & $00FF
      ;echo the incoming data back to the VB program
	tx_end = TXSTART;
   ;load beginning page for transmit buffer
    ;banksel(EWRPTL);
	EWRPTL = TXSTART & $00FF
	EWRPTH = (TXSTART & $FF00) >> 8

   ;write control byte
	EDATA = tx_control_byte 
	tx_end = tx_end + 1

	tx_end = tx_end + (((packetheader[rec_bytecnt_high] << 8) + packetheader[rec_bytecnt_low]) - 4);
	;banksel(ETXNDL);
	ETXNDL = tx_end & $00FF 
	ETXNDH = (tx_end & $FF00) >> 8 

	i_udp = 0
	while(tx_end)
		EDATA = packet[0 + i_udp]
		i_udp = i_udp + 1
		tx_end = tx_end - 1
	wend		

   ;send the contents of the transmit buffer onto the network
	if TXERIF then
	 TXERIF = 0 
     TXRST = 1 
   	 TXRST = 0 
	endif
  	TXRTS = 1
	while(1 == (ECON1 & ECON1_TXRTS))
	wend
	endif
return

tcp:
return

ripple_crc:
		if acc0 > $FF then
		 acc1 = acc1 + 1
		 acc0 = acc0 & $00FF
		endif
		if acc1 > $FF then
		 acc2 = acc2 + 1
		 acc1 = acc1 & $00FF
		endif
		if acc2 > $FF then
		 acc3 = acc3 + 1
		 acc2 = acc2 & $00FF
		endif
		if acc3 > $FF then
		 hserout[13,10,"CHECKSUM OVERFLOW ERROR",13,10]
		endif
		return
clrcrc:
	acc0 = 0
	acc1 = 0
	acc2 = 0
	acc3 = 0
	arg0 = 0
	arg1 = 0
	arg2 = 0
	arg3 = 0
	return

dhcp_state_engine:
return